module winex.dynloader;
/***
 * A Dynamic (shared library) loader  copyright 2003 <Mike Wynn>one_mad_alien@yahoo.co.uk
 * version 2 (1st May 2004) (changed to work with dmd 0.8x+, added the structs to make it even easier)
 *
 * There are 2 ways to use this
 *
 * slow way:
 *
 * ImportFunc!( linkage, result_t, param_t ) myextfunc;
 * static this() { myextfunc.use( "mydll.dll", "myfunc" ); }
 * // or ImportProc( .... ) if there is no return value
 *
 * fast way
 * instance DynFunc( linkage, result_t, param_t ) r_p_Ldr;
 * r_p_Ldr.fp myextfunc;
 * static this() { r_p_Ldr.use( myextfunc, "mydll.dll", "myfunc" ); }
 *
 * the difference is that there is an opCall over head for the slow version
 * the fast version is basically what the compiler will do implicitly if your
 * link against a lib created from the dll.
 * The slow version allows the types to only be declared once, and avoid the alias.
 *
 * from a users perspective there is no difference at all.
 *
 * in release+inline builds the fast version is about 5% faster.
 * in debug builds the fast version is about 8% faster.
 *
 */

import std.c.windows.windows;
//import std.c.stdio;

enum Linkage { Windows, C, D }
alias Linkage.Windows Windows;
alias Linkage.C C;
alias Linkage.D D;

version(Win32) {
class SharedLibrary {
	private HMODULE sysModule;
	protected this( char[] path ) {
		sysModule = LoadLibraryA( path );
		if ( !sysModule ) throw new Exception( "unable to load shared library " ~ path );
	}
	public FARPROC getFunctionAddress( char[] funcName )
	{
		FARPROC fp = GetProcAddress( sysModule, cast(char*)funcName );
		if ( !fp ) throw new Exception( "unable to get func address " ~ funcName );
		return fp;
	}
	private static SharedLibrary[char[]] loaded;
	public static SharedLibrary loadLibrary( char[] path )
	{
		SharedLibrary sl;
		if ( (sl = loaded[path]) === null ) return loaded[path] = new SharedLibrary( path );
		return sl;
	}
}
} else { // versoion(WINDOWS)
	// Not Implemented YET on other platforms.
	class SharedLibrary {
		public void function() getFunctionAddress( char[] funcName )
		{
			return null;
		}
		public static SharedLibrary loadLibrary( char[] path )
		{
			throw new Exception( "(NIY) unable to load shared library " ~ path );
			return null;
		}
	}
}

template Dyn( T ) {
	T require( char[] lib, char[] name ) {
		return cast(T)((SharedLibrary.loadLibrary( lib )).getFunctionAddress( name ));
	}
	void use( out T ef, char[] lib, char[] name ) {
		ef = require( lib, name );
	}
}

template DynProc( Linkage T : Linkage.D ) {
	void erf() { throw new Exception( "Shared Function not loaded" ); }
	typedef void function () fp = &erf;
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynProc( Linkage T : Linkage.Windows ) {
	extern ( Windows ) {
		void erf() { throw new Exception( "Shared Function not loaded" ); }
		typedef void function() fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynProc( Linkage T : Linkage.C ) {
	extern (C) {
		void erf() { throw new Exception( "Shared Function not loaded" ); }
		typedef void function() fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynProc( Linkage T : Linkage.D, P ) {
	void erf( P p ) { throw new Exception( "Shared Function not loaded" ); }
	typedef void function ( P ) fp = &erf;
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynProc( Linkage T : Linkage.Windows, P ) {
	extern (Windows) {
		void erf( P p ) { throw new Exception( "Shared Function not loaded" ); }
		typedef void function ( P ) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynProc( Linkage T : Linkage.C, P ) {
	extern ( C ) {
		void erf( P p ) { throw new Exception( "Shared Function not loaded" ); }
		typedef void function ( P ) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynProc( Linkage T : Linkage.D, P0, P1 ) {
	void erf( P0 p0, P1 p1 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
	typedef void function ( P0, P1 ) fp = &erf;
	alias Dyn!( fp ) dynFunc;

	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynProc( Linkage T : Linkage.Windows, P0, P1 ) {
	extern (Windows) {
		void erf( P0 p0, P1 p1 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
		typedef void function ( P0, P1 ) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynProc( Linkage T : Linkage.C, P0, P1 ) {
	extern ( C ) {
		void erf( P0 p0, P1 p1 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
		typedef void function ( P0, P1 ) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}


template DynFunc( Linkage T ) {
	alias DynProc!( T ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}

template DynFunc( Linkage T, R : void ) {
	alias DynProc!( T ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}

template DynFunc( Linkage T : Linkage.D, R ) {
	R erf() { throw new Exception( "Shared Function not loaded" ); return R.init; }
	typedef R function () fp = &erf;
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynFunc( Linkage T : Linkage.C, R ) {
	extern ( C ) {
		R erf() { throw new Exception( "Shared Function not loaded" ); return R.init; }
		typedef R function () fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynFunc( Linkage T : Linkage.Windows, R ) {
	extern ( Windows ) {
		R erf() { throw new Exception( "Shared Function not loaded" ); return R.init; }
        typedef R function () fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynFunc( Linkage T : Linkage.D, R : void, P ) {
	alias DynProc!( T, P ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}

template DynFunc( Linkage T : Linkage.C, R : void, P ) {
	alias DynProc!( T, P ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}

template DynFunc( Linkage T : Linkage.Windows, R : void, P ) {
	alias DynProc!( T, P ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}

template DynFunc( Linkage T : Linkage.D, R, P ) {
	R erf( P p ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
	typedef R function (P) fp = &erf;
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynFunc( Linkage T : Linkage.Windows, R, P ) {
	extern ( Windows ) {
		R erf( P p ) { 
            throw new Exception( "Shared Function not loaded" ); return R.init; 
        }
		typedef R function (P) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynFunc( Linkage T : Linkage.C, R, P ) {
	extern( C ) {
		R erf( P p ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
		typedef R function (P) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynFunc( Linkage T : Linkage.D, R : void, P0, P1 ) {
	alias DynProc!( T, P0, P1 ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}
template DynFunc( Linkage T : Linkage.Windows, R : void, P0, P1 ) {
	alias DynProc!( T, P0, P1 ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}
template DynFunc( Linkage T : Linkage.C, R : void, P0, P1 ) {
	alias DynProc!( T, P0, P1 ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}

template DynFunc( Linkage T : Linkage.D, R, P0, P1 ) {
	R erf( P0 p0, P1 p1 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
	typedef R function (P0,P1) fp = &erf;
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynFunc( Linkage T : Linkage.Windows, R, P0, P1 ) {
	extern( Windows ) {
		R erf( P0 p0, P1 p1 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
		typedef R function (P0,P1) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynFunc( Linkage T : Linkage.C, R, P0, P1 ) {
	extern ( C ) {
		R erf( P0 p0, P1 p1 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
		typedef R function (P0,P1) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynFunc( Linkage T : Linkage.D, R : void, P0, P1, P2 ) {
	alias DynProc!( T, P0, P1, P2 ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}
template DynFunc( Linkage T : Linkage.Windows, R : void, P0, P1, P2 ) {
	alias DynProc!( T, P0, P1, P2 ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}
template DynFunc( Linkage T : Linkage.C, R : void, P0, P1, P2 ) {
	alias DynProc!( T, P0, P1, P2 ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}
/*
template DynFunc( Linkage T, R : void, P0, P1, P2 ) {
	DynProc!( T, P0, P1, P2 ) realTemplate;
	alias realTemplate.erf erf;
	alias realTemplate.fp fp;
	alias realTemplate.use use;
	alias realTemplate.require require;
}
*/
template DynFunc( Linkage T : Linkage.D, R, P0, P1, P2 ) {
	R erf( P0 p0, P1 p1, P2 p2 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
	typedef R function (P0, P1, P2) fp = &erf;
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

template DynFunc( Linkage T : Linkage.Windows, R, P0, P1, P2 ) {
	extern ( Windows ) {
		R erf( P0 p0, P1 p1, P2 p2 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
		typedef R function (P0, P1, P2) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}
template DynFunc( Linkage T : Linkage.C, R, P0, P1, P2 ) {
	extern ( C ) {
		R erf( P0 p0, P1 p1, P2 p2 ) { throw new Exception( "Shared Function not loaded" ); return R.init; }
		typedef R function (P0, P1, P2) fp = &erf;
	}
	alias Dyn!( fp ) dynFunc;
	alias dynFunc.use use;
	alias dynFunc.require require;
}

struct ImportFunc( Linkage T, R )
{
public:
	alias DynFunc!( T, R ) loader;
private:
	loader.fp m_fp;
public:
	void use( char[] library, char[] name ) { loader.use( m_fp, library, name ); }
	R opCall() { return m_fp(); }
}

struct ImportFunc( Linkage T, R, P )
{
public:
	alias DynFunc!( T, R, P ) loader;
private:
	loader.fp m_fp;
public:
	void use( char[] library, char[] name ) { loader.use( m_fp, library, name ); }
	R opCall( P p ) { return m_fp( p ); }
}

struct ImportFunc( Linkage T, R, P0, P1 )
{
public:
	alias DynFunc!( T, R, P0, P1 ) loader;
private:
	loader.fp m_fp;
public:
	void use( char[] library, char[] name ) { loader.use( m_fp, library, name ); }
	R opCall( P0 p0, P1 p1 ) { return m_fp( p0, p1 ); }
}

struct ImportFunc( Linkage T, R, P0, P1, P2 )
{
public:
	alias DynFunc!( T, R, P0, P1, P2 ) loader;
private:
	loader.fp m_fp;
public:
	void use( char[] library, char[] name ) { loader.use( m_fp, library, name ); }
	R opCall( P0 p0, P1 p1, P2 p2 ) { return m_fp( p0, p1, p2 ); }
}

struct ImportProc( Linkage T, P )
{
public:
	alias DynProc!( T, P ) loader;
private:
	loader.fp m_fp;
public:
	void use( char[] library, char[] name ) { loader.use( m_fp, library, name ); }
	void opCall( P p ) { m_fp( p ); }
}

struct ImportProc( Linkage T, P0, P1 )
{
public:
	alias DynProc!( T, P0, P1 ) loader;
private:
	loader.fp m_fp;
public:
	void use( char[] library, char[] name ) { loader.use( m_fp, library, name ); }
	void opCall( P0 p0, P1 p1 ) { m_fp( p0, p1 ); }
}

struct ImportProc( Linkage T, P0, P1, P2 )
{
public:
	alias DynProc!( T, P0, P1, P2 ) loader;
private:
	loader.fp m_fp;
public:
	void use( char[] library, char[] name ) { loader.use( m_fp, library, name ); }
	void opCall( P0 p0, P1 p1, P2 p2 ) { m_fp( p0, p1, p2 ); }
}



